#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later

# depthcharge-tools depthchargectl program
# Copyright (C) 2021 Alper Nebi Yasak <alpernebiyasak@gmail.com>
# See COPYRIGHT and LICENSE files for full copyright information.

from depthcharge_tools.depthchargectl import depthchargectl

if __name__ == "__main__":
    depthchargectl.main()
